@@ -1,8 +1,5 @@
-import numpy as np
-from scipy.stats import multivariate_normal, norm, t, gamma, uniform, multivariate_t, logser
+from .libraries import *
 from . import helperfunctions as hf
-import time  # used for setting random number generator seed if None
-from twiggy import quick_setup, log
 
 quick_setup()
 logger = log.name('copulas')
@@ -39,6 +36,12 @@ class ClaytonCopula:
 
     @dim.setter
     def dim(self, value):
+        assert (value > 1), logger.error("Dimension must be > 1")
+        try:
+            value = int(value)
+        except Exception:
+            logger.error('Please provide dimension as an integer')
+            raise
         self.__dim = value
 
     def cdf(self, x):
@@ -78,15 +81,17 @@ class ClaytonCopula:
         """
         random_state = int(time.time()) if random_state is None else random_state
         assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
 
+        assert (size > 0), logger.error("Size must be > 0")
         try:
             size = int(size)
         except Exception:
             logger.error('Please make sure random_state is provided correctly')
             raise
 
-        gamma_sim = gamma.rvs(1 / self.par, size=[size, 1], random_state=random_state)
-        exp_sim = gamma.rvs(1, size=[size, self.dim], random_state=random_state + 2)
+        gamma_sim = stats.gamma.rvs(1 / self.par, size=[size, 1], random_state=random_state)
+        exp_sim = stats.gamma.rvs(1, size=[size, self.dim], random_state=random_state + 2)
         output = (1 + exp_sim / gamma_sim) ** (-1 / self.par)
         return output
 
@@ -122,7 +127,12 @@ class FrankCopula:
 
     @dim.setter
     def dim(self, value):
-        assert isinstance(value, int), logger.error('%r is not an integer' % value)
+        assert (value > 1), logger.error("Dimension must be > 1")
+        try:
+            value = int(value)
+        except Exception:
+            logger.error('Please provide dimension as an integer')
+            raise
         self.__dim = value
 
     def cdf(self, x):
@@ -166,15 +176,17 @@ class FrankCopula:
         """
         random_state = int(time.time()) if random_state is None else random_state
         assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
 
+        assert (size > 0), logger.error("Size must be > 0")
         try:
             size = int(size)
         except Exception:
             logger.error('Please make sure random_state is provided correctly')
             raise
 
-        logarithmic_sim = logser.rvs(1 - np.exp(-self.par), size=[size, 1], random_state=random_state)
-        exp_sim = gamma.rvs(1, size=[size, self.dim], random_state=random_state)
+        logarithmic_sim = stats.logser.rvs(1 - np.exp(-self.par), size=[size, 1], random_state=random_state)
+        exp_sim = stats.gamma.rvs(1, size=[size, self.dim], random_state=random_state)
         output = -1 / self.par * np.log(1 + np.exp(-exp_sim / logarithmic_sim) * (np.exp(-self.par) - 1))
         return output
 
@@ -210,7 +222,12 @@ class GumbelCopula:
 
     @dim.setter
     def dim(self, value):
-        assert isinstance(value, int), logger.error('%r is not an integer' % value)
+        assert (value > 1), logger.error("Dimension must be > 1")
+        try:
+            value = int(value)
+        except Exception:
+            logger.error('Please provide dimension as an integer')
+            raise
         self.__dim = value
 
     def cdf(self, x):
@@ -249,7 +266,9 @@ class GumbelCopula:
         """
         random_state = int(time.time()) if random_state is None else random_state
         assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
 
+        assert (size > 0), logger.error("Size must be > 0")
         try:
             size = int(size)
         except TypeError:
@@ -257,12 +276,12 @@ class GumbelCopula:
             raise
 
         a_ = 1 / self.par
-        uniform_sim = (uniform.rvs(size=[size, 1], random_state=random_state) - 0.5) * np.pi
-        exp_sim = gamma.rvs(1, size=[size, 1], random_state=random_state)
+        uniform_sim = (stats.uniform.rvs(size=[size, 1], random_state=random_state) - 0.5) * np.pi
+        exp_sim = stats.gamma.rvs(1, size=[size, 1], random_state=random_state)
         stable_sim = np.sin(a_ * (np.pi / 2 + uniform_sim)) / np.cos(
                     uniform_sim) ** (1 / a_) * (
                                  np.cos(uniform_sim - a_ * (np.pi / 2 + uniform_sim)) / exp_sim) ** ((1 - a_) / a_)
-        exp_sim = gamma.rvs(1, size=[size, self.dim], random_state=random_state + 2)
+        exp_sim = stats.gamma.rvs(1, size=[size, self.dim], random_state=random_state + 2)
         output = np.exp(-(exp_sim / stable_sim) ** (1 / self.par))
         return output
 
@@ -271,6 +290,7 @@ class GumbelCopula:
 class GaussCopula:
     """
     Gaussian copula.
+
     :param corr: Correlation matrix.
     :type corr: ``numpy.ndarray``
     """
@@ -305,7 +325,7 @@ class GaussCopula:
         :return: Cumulative distribution function in x.
         :rtype: ``numpy.ndarray``
         """
-        return multivariate_normal.cdf(norm.ppf(x), cov=self.corr)
+        return stats.multivariate_normal.cdf(stats.norm.ppf(x), cov=self.corr)
 
     def rvs(self, size=1, random_state=None):
         """
@@ -322,19 +342,20 @@ class GaussCopula:
         random_state = int(time.time()) if random_state is None else random_state
         assert isinstance(random_state, int), logger.error("%r has to be an integer" % random_state)
 
+        assert (size > 0), logger.error("Size must be > 0")
         try:
             size = int(size)
         except TypeError:
             logger.error('Please make sure random_state is provided correctly.')
             raise
 
-        sim = multivariate_normal.rvs(
+        sim = stats.multivariate_normal.rvs(
             mean=np.zeros(self.dim),
             cov=self.corr,
             size=size,
             random_state=random_state
         )
-        return norm.cdf(sim)
+        return stats.norm.cdf(sim)
 
 
 # TCopula
@@ -385,26 +406,98 @@ class TCopula:
     def error_cdf(self):
         return self.__error_cdf
 
-    def cdf(self, x, tolerance=1e-4, max_evaluations=1e+6, n_repetitions=30):
+    def cdf(self, x, tolerance=1e-4, n_iterations=30):
+        """
+        Cumulative distribution function.
+
+        :param x: quantile where the cumulative distribution function is evaluated.
+                    Array with shape (n, d) where n is the number of data points and d the dimension.
+        :type x: ``numpy.ndarray``
+        :param tolerance: tolerance threshold of approximation (default is 1e-4).
+        :type tolerance: ``float``, optional
+        :param n_iterations: number of iteration (default is 30).
+        :type n_iterations: ``int``, optional
+        
+        :return: Cumulative distribution function in x.
+        :rtype: ``numpy.ndarray``
+        """
+        q = stats.t(self.df).ppf(x)
+        (prob, err) = hf.multivariate_t_cdf(q, self.corr, self.df, tolerance, n_iterations)
+        self.__error_cdf = err
+        return prob
+
+    def rvs(self, size=1, random_state=None):
+        """
+        Random variates.
+
+        :param size: random variates sample size (default is 1).
+        :type size: ``int``, optional
+        :param random_state: random state for the random number generator.
+        :type random_state: ``int``, optional
+
+        :return: Random variates.
+        :rtype: ``numpy.float64`` or ``numpy.ndarray``
+        """
+        random_state = int(time.time()) if random_state is None else random_state
+        assert isinstance(random_state, int), logger.error("%r has to be an integer" % random_state)
+
+        assert (size > 0), logger.error("Size must be > 0")
+        try:
+            size = int(size)
+        except TypeError:
+            logger.error('Please make sure random_state is parametrized correctly.')
+            raise
+
+        sim = stats.multivariate_t.rvs(
+            df=self.df,
+            shape=self.corr,
+            size=size,
+            random_state=random_state
+        )
+        return stats.t.cdf(sim, df=self.df)
+
+# Independent
+class IndependentCopula:
+    """
+    Independent copula.
+
+    :param dim: copula dimension.
+    :type dim: ``int``
+    """
+
+    def __init__(self, dim):
+        self.dim = dim
+
+    @property
+    def dim(self):
+        return self.__dim
+
+    @dim.setter
+    def dim(self, value):
+        assert (value > 1), logger.error("Dimension must be > 1")
+        try:
+            value = int(value)
+        except Exception:
+            logger.error('Please provide dimension as an integer')
+            raise
+        self.__dim = value
+
+    def cdf(self, x):
         """
         Cumulative distribution function.
 
         :param x: Array with shape (N, d) where N is the number of points and d the dimension.
         :type x: ``numpy.ndarray``
-        :param tolerance: Tolerance threshold in cdf computation.
-        :type tolerance: ``float``
-        :param max_evaluations: Maximum number of iterations.
-        :type max_evaluations: ``int``
-        :param n_repetitions: Array with shape (N, d) where N is the number of points and d the dimension.
-        :type n_repetitions: ``int``
-        
         :return: Cumulative distribution function in x.
         :rtype: ``numpy.ndarray``
         """
-        q = t(self.df).ppf(x)
-        (prob, err) = hf.multivariate_t_cdf(q, self.corr, self.df, tolerance, max_evaluations, n_repetitions)
-        self.__error_cdf = err
-        return prob
+        try:
+            x = x.reshape(-1, self.dim)
+        except Exception:
+            logger.error('Please make sure x dimension is the same as copula dimension')
+            raise
+
+        return np.prod(x, axis=1)
 
     def rvs(self, size=1, random_state=None):
         """
@@ -419,18 +512,137 @@ class TCopula:
         :rtype: ``numpy.float64`` or ``numpy.ndarray``
         """
         random_state = int(time.time()) if random_state is None else random_state
-        assert isinstance(random_state, int), logger.error("%r has to be an integer" % random_state)
+        assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
 
+        assert (size > 0), logger.error("Size must be > 0")
         try:
             size = int(size)
-        except TypeError:
-            logger.error('Please make sure random_state is parametrized correctly.')
+        except Exception:
+            logger.error('Please make sure random_state is provided correctly')
             raise
 
-        sim = multivariate_t.rvs(
-            df=self.df,
-            shape=self.corr,
-            size=size,
-            random_state=random_state
-        )
-        return t.cdf(sim, df=self.df)
+        return np.random.uniform(size=(size, self.dim))
+
+# Fréchet–Hoeffding Lower Bound 
+class FHLowerCopula:
+    """
+    Fréchet–Hoeffding lower bound bidimensional copula.
+    """
+
+    def __init__(self):
+        pass
+
+    @staticmethod
+    def dim():
+        return 2
+
+    def cdf(self, x):
+        """
+        Cumulative distribution function.
+
+        :param x: Array with shape (N, 2) where N is the number of points.
+        :type x: ``numpy.ndarray``
+        :return: Cumulative distribution function in x.
+        :rtype: ``numpy.ndarray``
+        """
+        try:
+            x = x.reshape(-1, 2)
+            logger.warning('x shape second dimension set to 2')
+        except Exception:
+            logger.error('Please make sure x shape second dimension is 2')
+            raise
+
+        return np.maximum(np.sum(x, axis=1) - 1, 0)
+
+    def rvs(self, size=1, random_state=None):
+        """
+        Random variates.
+
+        :param size: random variates sample size (default is 1).
+        :type size: ``int``
+        :param random_state: random state for the random number generator.
+        :type random_state: ``int``
+
+        :return: Random variates.
+        :rtype: ``numpy.float64`` or ``numpy.ndarray``
+        """
+        random_state = int(time.time()) if random_state is None else random_state
+        assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
+
+        assert (size > 0), logger.error("Size must be > 0")
+        try:
+            size = int(size)
+        except Exception:
+            logger.error('Please make sure random_state is provided correctly')
+            raise
+
+        u = np.random.uniform(size=(size, 1))
+        return np.concatenate((u, 1-u), axis=1)
+
+# Fréchet–Hoeffding Upper Bound 
+class FHUpperCopula:
+    """
+    Fréchet–Hoeffding upper bound copula.
+    """
+
+    def __init__(self, dim):
+        self.dim = dim
+
+    @property
+    def dim(self):
+        return self.__dim
+
+    @dim.setter
+    def dim(self, value):
+        assert (value > 1), logger.error("Dimension must be > 1")
+        try:
+            value = int(value)
+        except Exception:
+            logger.error('Please provide dimension as an integer')
+            raise
+        self.__dim = value
+
+    def cdf(self, x):
+        """
+        Cumulative distribution function.
+
+        :param x: Array with shape (N, d) where N is the number of points and d the dimension.
+        :type x: ``numpy.ndarray``
+        :return: Cumulative distribution function in x.
+        :rtype: ``numpy.ndarray``
+        """
+        try:
+            x = x.reshape(-1, self.dim)
+        except Exception:
+            logger.error('Please make sure x dimension is the same as copula dimension')
+            raise
+
+        return np.min(x, axis=1)
+
+    def rvs(self, size=1, random_state=None):
+        """
+        Random variates.
+
+        :param size: random variates sample size (default is 1).
+        :type size: ``int``
+        :param random_state: random state for the random number generator.
+        :type random_state: ``int``
+
+        :return: Random variates.
+        :rtype: ``numpy.float64`` or ``numpy.ndarray``
+        """
+        random_state = int(time.time()) if random_state is None else random_state
+        assert isinstance(random_state, int), logger.error("%r is not an integer" % random_state)
+        np.random.seed(random_state)
+
+        assert (size > 0), logger.error("Size must be > 0")
+        try:
+            size = int(size)
+        except Exception:
+            logger.error('Please make sure random_state is provided correctly')
+            raise
+
+        u = np.random.uniform(size=(size, 1))
+        return np.tile(u, (1, self.dim))
\ No newline at end of file
